import 'dart:async';
import 'dart:convert';
import 'package:build/build.dart';
import 'package:recase/recase.dart';
import 'package:dart_style/dart_style.dart';

class ApiSpecBuilder implements Builder {
  @override
  Map<String, List<String>> get buildExtensions => {
        '.json': [
          '.models.dart',
          '.enums.dart',
          '.endpoints.dart',
          '.service.dart',
          '.repository.dart'
        ]
      };

  @override
  FutureOr<void> build(BuildStep buildStep) async {
    // Retrieve the currently matched asset
    AssetId inputId = buildStep.inputId;

    String inputName = inputId.path.split("/").last.replaceAll(".json", "");

    // Create a new target
    AssetId modelsAssetId = inputId.changeExtension('.models.dart');
    AssetId enumsAssetId = inputId.changeExtension('.enums.dart');
    AssetId endpointsAssetId = inputId.changeExtension('.endpoints.dart');
    AssetId serviceAssetId = inputId.changeExtension('.service.dart');
    AssetId abstractServiceAssetId =
        inputId.changeExtension('.repository.dart');

    final classBuffer = StringBuffer();
    final enumBuffer = StringBuffer();
    final endpointsBuffer = StringBuffer();
    final serviceBuffer = StringBuffer();
    final repositoryBuffer = StringBuffer();
    final formatter = DartFormatter();
    var contents = await buildStep.readAsString(inputId);
    var jsonData = await jsonDecode(contents);

    classBuffer
        .writeln("import 'package:json_annotation/json_annotation.dart';");
    classBuffer.writeln("import '$inputName.enums.dart';");
    classBuffer.writeln("\npart '$inputName.g.dart';");

    enumBuffer
        .writeln("import 'package:json_annotation/json_annotation.dart';");

    generateDefinitions(jsonData, classBuffer, enumBuffer);
    generateParameters(jsonData, classBuffer, enumBuffer);
    generateResponses(jsonData, classBuffer, enumBuffer);
    generatePathEndpoints(jsonData, endpointsBuffer);
    generateService(jsonData, serviceBuffer, inputName);
    generateAbstractService(jsonData, repositoryBuffer, inputName);

    await buildStep.writeAsString(
        modelsAssetId, formatter.format(classBuffer.toString()));
    await buildStep.writeAsString(
        enumsAssetId, formatter.format(enumBuffer.toString()));
    await buildStep.writeAsString(
        endpointsAssetId, formatter.format(endpointsBuffer.toString()));
    await buildStep.writeAsString(
        serviceAssetId, formatter.format(serviceBuffer.toString()));
    await buildStep.writeAsString(
        abstractServiceAssetId, formatter.format(repositoryBuffer.toString()));
  }

  void generateDefinitions(
      dynamic jsonData, StringBuffer classBuffer, StringBuffer enumBuffer) {
    dynamic definitions = jsonData["definitions"];
    (definitions ?? {}).forEach((definition, definitionValue) {
      if (definitionValue["allOf"] != null) {
        generateModelFromAllOf(
            jsonData, classBuffer, enumBuffer, definition, definitionValue);
      } else if (definitionValue["\$ref"] != null) {
        generateModelFromReference(
            jsonData, classBuffer, enumBuffer, definition, definitionValue);
      } else if (definitionValue["properties"] != null) {
        generateModelFromObject(
            jsonData, classBuffer, enumBuffer, definition, definitionValue);
      }
    });
  }

  void generateParameters(
    dynamic jsonData,
    StringBuffer classBuffer,
    StringBuffer enumBuffer,
  ) {
    dynamic parameters = jsonData["parameters"];
    (parameters ?? {}).forEach((parameter, parameterValue) {
      if (parameterValue["in"] != "body") {
        return;
      }
      if (parameterValue["schema"]["allOf"] != null) {
        generateModelFromAllOf(
            jsonData, classBuffer, enumBuffer, parameter, parameterValue);
      } else if (parameterValue["schema"]["\$ref"] != null) {
        generateModelFromReference(
            jsonData, classBuffer, enumBuffer, parameter, parameterValue);
      } else if (parameterValue["schema"]["properties"] != null) {
        generateModelFromObject(
            jsonData, classBuffer, enumBuffer, parameter, parameterValue);
      }
    });
  }

  void generateResponses(
    dynamic jsonData,
    StringBuffer classBuffer,
    StringBuffer enumBuffer,
  ) {
    dynamic responses = jsonData["responses"];
    (responses ?? {}).forEach((response, responseValue) {
      if (responseValue["schema"]["allOf"] != null) {
        generateModelFromAllOf(
            jsonData, classBuffer, enumBuffer, response, responseValue);
      } else if (responseValue["schema"]["\$ref"] != null) {
        generateModelFromReference(
            jsonData, classBuffer, enumBuffer, response, responseValue);
      } else if (responseValue["schema"]["properties"] != null) {
        generateModelFromObject(
            jsonData, classBuffer, enumBuffer, response, responseValue);
      }
    });
  }

  void generatePropertiesString(
      String propertyKey,
      dynamic propertyValue,
      List requiredProperties,
      StringBuffer classBuffer,
      StringBuffer enumBuffer) {
    String questionMarkInput =
        requiredProperties.contains(propertyKey) ? '' : '?';

    if (propertyValue["type"] == "string") {
      if (propertyValue.containsKey("enum")) {
        classBuffer.writeln(
            "${ReCase(propertyKey).pascalCase}Enum$questionMarkInput ${ReCase(propertyKey).camelCase};");
        generateEnum(propertyKey, propertyValue["enum"], enumBuffer);
      } else {
        classBuffer.writeln(
            'String$questionMarkInput ${ReCase(propertyKey).camelCase};');
      }
    } else if (propertyValue["type"] == "integer") {
      classBuffer
          .writeln('int$questionMarkInput ${ReCase(propertyKey).camelCase};');
    } else if (propertyValue["type"] == "boolean") {
      classBuffer
          .writeln('bool$questionMarkInput ${ReCase(propertyKey).camelCase};');
    } else if (propertyValue["type"] == "array") {
      if (propertyValue["items"]["type"] == "string") {
        if (propertyValue["items"].containsKey("enum")) {
          classBuffer.writeln(
              "List<${ReCase(propertyKey).pascalCase}Enum>$questionMarkInput ${ReCase(propertyKey).camelCase};");
          generateEnum(propertyKey, propertyValue["items"]["enum"], enumBuffer);
        } else {
          throw Exception(["$propertyKey not handled properly"]);
        }
      } else if (propertyValue["items"]["\$ref"] != null) {
        String reference = propertyValue["items"]["\$ref"].split('/').last;
        classBuffer.writeln(
            "List<$reference>$questionMarkInput ${ReCase(propertyKey).camelCase};");
      }
    } else if (propertyValue.containsKey("\$ref")) {
      String reference = propertyValue["\$ref"].split('/').last;
      classBuffer.writeln(
          "$reference$questionMarkInput ${ReCase(propertyKey).camelCase};");
    } else {
      throw Exception(["$propertyKey not handled properly"]);
    }
  }

  void generateEnum(
      String propertyKey, List enumList, StringBuffer enumBuffer) {
    // DON't Generate Enum if it already exists, may need to update if there
    // are two enums with same property name but different keys
    if (enumBuffer
        .toString()
        .contains("enum ${ReCase(propertyKey).pascalCase}Enum")) {
      return;
    }
    enumBuffer.writeln("enum ${ReCase(propertyKey).pascalCase}Enum {");
    for (final enumString in enumList) {
      enumBuffer.writeln('@JsonValue("$enumString")');
      enumBuffer.writeln('${ReCase(enumString).camelCase},');
    }
    enumBuffer.writeln("}\n");
  }

  void generateToAndFromJsonMap(String definition, StringBuffer classBuffer) {
    classBuffer.writeln(
        "\nfactory $definition.fromJson(Map<String, dynamic> json) =>_\$${definition}FromJson(json);");
    classBuffer.writeln(
        "\nMap<String, dynamic> toJson() => _\$${definition}ToJson(this);");
  }

  List getAllRequiredProperties(dynamic jsonData, dynamic allOfList) {
    List requiredProperties = [];
    for (final dictionary in allOfList) {
      if (dictionary["\$ref"] != null) {
        List referenceRequiredProperties =
            getReferenceRequiredProperties(jsonData, dictionary["\$ref"]);
        requiredProperties.addAll(referenceRequiredProperties);
      } else if (dictionary["properties"] != null &&
          dictionary.containsKey("required")) {
        requiredProperties.addAll(dictionary["required"]);
      }
    }
    return requiredProperties;
  }

  List getReferenceRequiredProperties(dynamic jsonData, String reference) {
    List referencePathSegments = reference.split('/');
    dynamic referenceObject =
        jsonData[referencePathSegments[1]][referencePathSegments[2]];
    List requiredProperties = [];

    if (referencePathSegments[1] == "definitions") {
      List definitionRequiredProperties =
          referenceObject.containsKey("required")
              ? referenceObject["required"]
              : [];
      requiredProperties.addAll(definitionRequiredProperties);
    }
    return requiredProperties;
  }

  void generateConstructorProperties(
      StringBuffer classBuffer, List requiredProperties, List objectKeys) {
    for (String property in objectKeys) {
      if (requiredProperties.contains(property)) {
        classBuffer.writeln("required this.${ReCase(property).camelCase},");
      } else {
        classBuffer.writeln("this.${ReCase(property).camelCase},");
      }
    }
  }

  void generateModelFromAllOf(dynamic jsonData, StringBuffer classBuffer,
      StringBuffer enumBuffer, String model, dynamic modelValue) {
    List allOfList = [];
    if (modelValue.containsKey("schema")) {
      allOfList = modelValue["schema"]["allOf"] ?? [];
    } else {
      allOfList = modelValue["allOf"] ?? [];
    }

    classBuffer.writeln('@JsonSerializable(fieldRename: FieldRename.snake)');
    classBuffer.writeln('class $model {');

    List requiredProperties = getAllRequiredProperties(jsonData, allOfList);

    // Properties generation
    for (dynamic dictionary in allOfList) {
      if (dictionary.containsKey("\$ref")) {
        List referencePathSegments = dictionary["\$ref"].split('/');
        dynamic referenceObject =
            jsonData[referencePathSegments[1]][referencePathSegments[2]];
        referenceObject["properties"].forEach((propertyKey, propertyValue) {
          generatePropertiesString(propertyKey, propertyValue,
              requiredProperties, classBuffer, enumBuffer);
        });
      } else if (dictionary.containsKey("properties")) {
        dictionary["properties"].forEach((propertyKey, propertyValue) {
          generatePropertiesString(propertyKey, propertyValue,
              requiredProperties, classBuffer, enumBuffer);
        });
      } else {
        throw Exception(["Unhandled $model"]);
      }
    }

    // generate Constructor
    classBuffer.writeln("\n");
    classBuffer.writeln("$model({");

    for (dynamic dictionary in allOfList) {
      if (dictionary.containsKey("\$ref")) {
        List referencePathSegments = dictionary["\$ref"].split('/');
        dynamic referenceObject =
            jsonData[referencePathSegments[1]][referencePathSegments[2]];

        generateConstructorProperties(classBuffer, requiredProperties,
            referenceObject["properties"].keys.toList());
      } else if (dictionary.containsKey("properties")) {
        generateConstructorProperties(classBuffer, requiredProperties,
            dictionary["properties"].keys.toList());
      } else {
        throw Exception(["Unhandled constructor $model"]);
      }
    }
    classBuffer.writeln("});");
    generateToAndFromJsonMap(model, classBuffer);
    classBuffer.writeln('}');
  }

  void generateModelFromReference(dynamic jsonData, StringBuffer classBuffer,
      StringBuffer enumBuffer, String model, dynamic modelValue) {
    String reference;
    if (modelValue.containsKey("schema")) {
      reference = modelValue["schema"]["\$ref"];
    } else {
      reference = modelValue["\$ref"];
    }

    classBuffer.writeln('@JsonSerializable(fieldRename: FieldRename.snake)');
    classBuffer.writeln('class $model {');

    List referencePathSegments = reference.split('/');
    dynamic referenceObject =
        jsonData[referencePathSegments[1]][referencePathSegments[2]];
    List requiredProperties =
        getReferenceRequiredProperties(jsonData, reference);
    referenceObject["properties"].forEach((propertyKey, propertyValue) {
      generatePropertiesString(propertyKey, propertyValue, requiredProperties,
          classBuffer, enumBuffer);
    });

    classBuffer.writeln("\n");
    classBuffer.writeln("$model({");
    generateConstructorProperties(classBuffer, requiredProperties,
        referenceObject["properties"].keys.toList());

    classBuffer.writeln("});");
    generateToAndFromJsonMap(model, classBuffer);
    classBuffer.writeln('}');
  }

  void generateModelFromObject(dynamic jsonData, StringBuffer classBuffer,
      StringBuffer enumBuffer, String model, dynamic modelValue) {
    dynamic modelValueObject;
    if (modelValue.containsKey("schema")) {
      modelValueObject = modelValue["schema"];
    } else {
      modelValueObject = modelValue;
    }
    List requiredProperties = (modelValueObject.containsKey("required"))
        ? modelValueObject["required"]
        : [];

    classBuffer.writeln('@JsonSerializable(fieldRename: FieldRename.snake)');
    classBuffer.writeln('class $model {');
    (modelValueObject["properties"]).forEach((propertyKey, propertyValue) {
      generatePropertiesString(propertyKey, propertyValue, requiredProperties,
          classBuffer, enumBuffer);
    });

    classBuffer.writeln("\n");
    classBuffer.writeln("$model({");

    generateConstructorProperties(classBuffer, requiredProperties,
        modelValueObject["properties"].keys.toList());

    classBuffer.writeln("});");
    generateToAndFromJsonMap(model, classBuffer);
    classBuffer.writeln('}');
  }

  void generatePathEndpoints(dynamic jsonData, StringBuffer endpointsBuffer) {
    dynamic paths = jsonData["paths"];
    endpointsBuffer.writeln("class Endpoints {");
    (paths ?? {}).forEach((path, pathValue) {
      (pathValue ?? {}).forEach((requestMethod, requestMethodValue) {
        endpointsBuffer.writeln(
            "static const String ${ReCase(requestMethodValue["operationId"]).camelCase} = \"$path\";");
      });
    });
    endpointsBuffer.writeln("}");
  }

  void generateService(
      dynamic jsonData, StringBuffer serviceBuffer, String inputName) {
    dynamic info = jsonData["info"] ?? {};
    String serviceName = info["title"] ?? "Service";
    serviceBuffer
        .writeln("import 'package:retrofit/retrofit.dart' as retrofit;");
    serviceBuffer.writeln("import 'package:dio/dio.dart';");
    serviceBuffer.writeln("import '$inputName.models.dart';");
    serviceBuffer.writeln("import '$inputName.endpoints.dart';");
    serviceBuffer.writeln("import '$inputName.repository.dart';");
    serviceBuffer.writeln("\npart '$inputName.service.g.dart';");

    serviceBuffer.writeln("@retrofit.RestApi()");

    serviceName = "${ReCase(serviceName).pascalCase}Service";
    serviceBuffer.writeln(
        "abstract class ${serviceName}Impl implements I$serviceName {");
    serviceBuffer
        .writeln("factory ${serviceName}Impl(Dio dio) = _${serviceName}Impl;");
    dynamic paths = jsonData["paths"];
    (paths ?? {}).forEach((path, pathValue) {
      (pathValue ?? {}).forEach((requestMethod, requestMethodValue) {
        String apiName = ReCase(requestMethodValue["operationId"]).camelCase;
        dynamic successResponse = requestMethodValue["responses"]["200"];
        List parameters = requestMethodValue["parameters"];
        String returnType = successResponse.containsKey("\$ref")
            ? successResponse["\$ref"].split("/")[2]
            : "void";

        serviceBuffer.writeln("@override");
        serviceBuffer.writeln(
            "@retrofit.${requestMethod.toUpperCase()}(Endpoints.$apiName)");

        serviceBuffer.writeln("Future<$returnType> $apiName(");
        for (final parameter in parameters) {
          List referencePathSegments = parameter["\$ref"].split('/');
          dynamic referenceObject =
              jsonData[referencePathSegments[1]][referencePathSegments[2]];
          if (referenceObject["in"] == "body") {
            String isRequired = referenceObject["required"] ? '' : '?';
            serviceBuffer.writeln(
                "@retrofit.Body() ${referencePathSegments[2]}$isRequired ${ReCase(referencePathSegments[2]).camelCase},");
          }
        }
        serviceBuffer.writeln(");");
        serviceBuffer.writeln("");
      });
    });
    serviceBuffer.writeln("}");
  }

  void generateAbstractService(
      dynamic jsonData, StringBuffer repositoryBuffer, String inputName) {
    dynamic info = jsonData["info"] ?? {};
    String serviceName = info["title"] ?? "Service";
    repositoryBuffer.writeln("import '$inputName.models.dart';");
    serviceName = "${ReCase(serviceName).pascalCase}Service";
    repositoryBuffer.writeln("abstract class I$serviceName {");
    dynamic paths = jsonData["paths"];
    (paths ?? {}).forEach((path, pathValue) {
      (pathValue ?? {}).forEach((requestMethod, requestMethodValue) {
        String apiName = ReCase(requestMethodValue["operationId"]).camelCase;
        dynamic successResponse = requestMethodValue["responses"]["200"];
        List parameters = requestMethodValue["parameters"];
        String returnType = successResponse.containsKey("\$ref")
            ? successResponse["\$ref"].split("/")[2]
            : "void";

        repositoryBuffer.writeln("Future<$returnType> $apiName(");
        for (final parameter in parameters) {
          List referencePathSegments = parameter["\$ref"].split('/');
          dynamic referenceObject =
              jsonData[referencePathSegments[1]][referencePathSegments[2]];
          if (referenceObject["in"] == "body") {
            String isRequired = referenceObject["required"] ? '' : '?';
            repositoryBuffer.writeln(
                "${referencePathSegments[2]}$isRequired ${ReCase(referencePathSegments[2]).camelCase},");
          }
        }
        repositoryBuffer.writeln(");");
        repositoryBuffer.writeln("");
      });
    });
    repositoryBuffer.writeln("}");
  }
}
